#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/core/utility.hpp>
#include <opencv2/saliency.hpp>
#include <opencv2/video/background_segm.hpp>
#include <opencv2/bgsegm.hpp>
#include <opencv2/optflow.hpp>
#include <opencv2/objdetect/objdetect.hpp>
#include <opencv2/video/tracking.hpp>

#include <opencv2/core/cuda.hpp>
#include <opencv2/cudafilters.hpp>
#include <opencv2/cudaimgproc.hpp>

#include <iostream>
#include <stdio.h>
#include <vector>
#include <map>
#include <list>
#include <sstream>
#include <fstream>
#include <regex>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <memory>
#include <time.h>
#include <cmath>
#include <random>

using namespace std;
using namespace cv;
using namespace saliency;

// Log size for method 'framecnt'
#define FRAME_LOG_SIZE 5

/*
Valid methods:
	- motionseg, framecnt, optflow
*/
static const char* keys =
{   "{@video_name      | | video name            }"
    "{@start_frame     |1| Start frame           }"
    "{@hold_frame	   |2| Hold frame 			 }"
    "{@method  		   |3| method 			     }"
    "{@optflow_sub	   |4| Optflow submethod     }"
};

string g_videoName;

/*	Computes intensity histogram of src, and stores in it dst.
	Parameters:
		otsoThresh:		If non-zero, a red vertical line will be drawn to mark it
		mask:			If non-zero, histogram computation will be done on src with mask applied
*/
void histogramCompute(const Mat& src, Mat& dst, double otsuThresh = 0, const Mat& mask = Mat(), bool returnRawHistograms = false)
{
	cout << "Thresh: " << otsuThresh << endl;
	Mat histData(Size(1,src.rows*src.cols+1), CV_8UC1);
	for (int i=0; i<src.rows; i++)
	{
		for (int j=0; j<src.cols; j++)
		{
			histData.at<char>(0,i*src.cols + j) = src.at<char>(i,j);
		}
	}

	Mat hist;
	int histSize = 256;
	float range[] = { 0, 256 } ;
	const float* histRange = { range };
	int hist_w = 512; 
	int hist_h = 400;
	int bin_w = cvRound( (double) hist_w/histSize );

	calcHist(&histData, 1, 0, mask, hist, 1, &histSize, &histRange, true, false);

	Mat histImage (hist_h, hist_w, CV_8UC3, Scalar(0,0,0));
	normalize(hist, hist, 0, histImage.rows, NORM_MINMAX, -1, Mat());

	for( int i = 1; i < histSize; i++ )
	{
		line( histImage, Point( bin_w*(i-1), hist_h - cvRound(hist.at<float>(i-1)) ) ,
						 Point( bin_w*(i), hist_h - cvRound(hist.at<float>(i)) ),
						 Scalar( 255, 0, 0), 2, 8, 0 );
		if (i == round(otsuThresh))
		{
			line( histImage, Point( bin_w*(i), hist_h ) ,
						 Point( bin_w*(i), 0 ),
						 Scalar( 0, 0, 255), 2, 8, 0 );
		}
	}

	histImage.copyTo(dst);

	histData.release();
	hist.release();
	histImage.release();
}


int main( int argc, char** argv ) {

	// Parse command line input
	CommandLineParser parser( argc, argv, keys );
	g_videoName = parser.get<String>( 0 );
	int startFrame = parser.get<int>( 1 );
	int holdFrame = parser.get<int>( 2 );
	string PROCESS_METHOD = parser.get<string>( 3 );
	string optflow_submethod = parser.get<String>( 4 );

	// Init video capture
	VideoCapture cap;
	cap.open(g_videoName);
	if( !cap.isOpened() ) {
		cout << "Failed to open video: " << g_videoName << endl;
	}
	cap.set(CAP_PROP_POS_FRAMES, static_cast<int>( startFrame ));

	Mat frame, image;
	Size displaySize(1066,600);

	// Init saliency
	Ptr<Saliency> saliencyAlgorithm;	
	Mat saliencyMap, saliencyBinary, saliencyProcessed, salBinTemp, saliencyBinary_segmented;

	// Method specific variables
	Mat mhi(displaySize, CV_32FC1), segmask;
	list<Mat> frame_log(FRAME_LOG_SIZE, Mat(displaySize, CV_8UC1, Scalar(0,0,0)));
	//// Optflow
	Mat flow;
	UMat flowUMat, prevgray;
	Mat flowProcessResult(frame.rows, frame.cols, CV_8UC1), flowMagnitude(frame.rows, frame.cols, CV_8UC1);
	
	clock_t last = clock();
	unsigned int frame_num = startFrame;
	
	while (1) {
		// Read a frame from the video
		cap >> frame;
		if( frame.empty() ) {
			cout << "Frame empty" << endl;
			return 0;
		}

		// Resize the frame to make it faster to work with
		resize(frame, frame, displaySize);
		cvtColor(frame, image, COLOR_BGR2GRAY);

		// Init saliency processing
		if (!saliencyAlgorithm)
		{
			saliencyAlgorithm = MotionSaliencyBinWangApr2014::create();
			saliencyAlgorithm.dynamicCast<MotionSaliencyBinWangApr2014>()->setImagesize( image.cols, image.rows );
			saliencyAlgorithm.dynamicCast<MotionSaliencyBinWangApr2014>()->init();
		}

		
		if( saliencyAlgorithm->computeSaliency( image, saliencyMap ) )
		{
			saliencyMap.copyTo(saliencyBinary);

			// Open binary saliency map with a square
			/*cuda::GpuMat saliencyBinaryGpu;
			saliencyBinaryGpu.upload(saliencyMap);

			// Noise filter
			int kernel_type = MORPH_RECT;
			int kernel_size = 2;
			Mat kernel = getStructuringElement( kernel_type,
		                                   Size( 2*kernel_size + 1, 2*kernel_size+1 ),
		                                   Point( kernel_size, kernel_size ) );

			Ptr<cuda::Filter> filter = cuda::createMorphologyFilter(MORPH_OPEN, saliencyBinaryGpu.type(), kernel);
			filter->apply(saliencyBinaryGpu, saliencyBinaryGpu);

			saliencyBinaryGpu.download(saliencyBinary);
			*/
			if (PROCESS_METHOD == "motionseg")
			{
				//Motion segmentation
				saliencyBinary_segmented = saliencyBinary.clone();
				clock_t now = clock();
				double now_millis = static_cast<double>(now)/CLOCKS_PER_SEC;
				double interval = static_cast<double>(now - last)/CLOCKS_PER_SEC;
		    	motempl::updateMotionHistory(saliencyBinary_segmented, mhi, now_millis, interval*10);
		    	vector<Rect> boundingRects;
		    	motempl::segmentMotion(mhi, segmask, boundingRects, now_millis, interval*2);
		    	last = clock();
		    	for (auto it=boundingRects.begin(); it != boundingRects.end(); ++it)
		    	{
		    		rectangle(saliencyBinary_segmented, *it, Scalar(255,0,255));
		    	}

		    	//Contour drawing
			    vector<vector<Point>> contours;
			    vector<Vec4i> hierarchy;
			    findContours( saliencyBinary, contours, hierarchy, CV_RETR_CCOMP, CV_CHAIN_APPROX_SIMPLE );
		    	for (auto it=contours.begin(); it != contours.end(); ++it)
		    	{
		    		rectangle(saliencyBinary, boundingRect(*it), Scalar(255,0,255));
		    	}
		    }
		    else if (PROCESS_METHOD == "framecnt")
		    {
		    	Mat inverseBinary, show;
		    	threshold(saliencyBinary, inverseBinary, 128, 255, THRESH_BINARY_INV);
		    	frame_log.push_back(inverseBinary);
		    	frame_log.pop_front();

		    	auto it = frame_log.begin();
		    	Mat mask = *it;
		    	for (++it; it!=frame_log.end(); ++it)
		    	{
		    		bitwise_or(mask, *it, mask);
		    	}

		    	bitwise_and(mask, saliencyBinary, show);
		    	imshow("framecnt", show);
		    }
		    else if (PROCESS_METHOD == "optflow")
		    {
		    	imshow("saliencyBinary pre optflow", saliencyBinary);
		    	if ( prevgray.empty() == false )
		    	{
		    		calcOpticalFlowFarneback(prevgray, image, flowUMat, 0.4, 1, 12, 2, 8, 1.2, 0);
		    		flowUMat.copyTo(flow);

		    		Mat flowTemp(frame.rows, frame.cols, CV_64FC1);
		    		for (int y=0; y<frame.rows; y++)
		    		{
		    			for (int x=0; x<frame.cols; x++)
		    			{
		    				const Point2f flowatxy = flow.at<Point2f>(y,x);
		    				double mag = sqrt(pow(flowatxy.x,2) + pow(flowatxy.y,2));
		    				flowTemp.at<double>(y,x) = mag;
		    				
		    				if (mag < 1)
		    				{
		    					saliencyBinary.at<char>(y,x) = 0;
		    				}
		    			}
		    		}

		    		double min, max;
		    		minMaxLoc(flowTemp, &min, &max);
		    		flowTemp = flowTemp*256/max;
		    		flowTemp.convertTo(flowMagnitude, CV_8UC1);

		    		Mat flowHist;

		    		if (optflow_submethod == "otsu")
		    		{
			    		GaussianBlur(flowMagnitude, flowMagnitude, Size(5,5), 0.3*((5-1)*0.5 - 1) + 0.8);
			    		double otsuThresh = threshold(flowMagnitude, flowProcessResult, 0, 255, THRESH_BINARY | THRESH_OTSU);

			    		// Create flow magnitude histogram
			    		minMaxLoc(flowMagnitude, &min, &max);
			    		histogramCompute(flowMagnitude*256/max, flowHist, otsuThresh*256/max);
			    		imshow("Flow histogram", flowHist);
			    	}
			    	else {
			    		adaptiveThreshold(flowMagnitude, flowProcessResult, 255, ADAPTIVE_THRESH_GAUSSIAN_C, THRESH_BINARY, 501, -10);
			    	}

		    		Mat tmpO, tmpB;
		    		frame.copyTo(tmpO, flowProcessResult);
		    		imshow("Optflow processed result", flowProcessResult);
					imshow("Flow magnitude", flowMagnitude);
		    	}
		    	
		    	image.copyTo(prevgray);

		    }
		}

		imshow("Original", image);
		imshow("saliencyMap", saliencyMap);
		imshow("saliencyBinary", saliencyBinary);
		//imshow("Segmask", segmask);

		char k = static_cast<char>(waitKey(1));
		if ( k == 'q') { break; }
		else if ( (k == 'p') || (++frame_num == holdFrame) ) 
		{
			cout << "Frame: " << frame_num << endl;
			char c = static_cast<char>(waitKey(0));
			if ( c == 'q' ) 
			{ 
				break; 
			}
			else if (c == 's') 
			{
					cout << "Storing images" << endl;

					vector<int> compression_params;
				    compression_params.push_back(CV_IMWRITE_PNG_COMPRESSION);
				    compression_params.push_back(3);

					imwrite("<INSERT IMAGE PATH>", frame, compression_params);
					imwrite("<INSERT IMAGE PATH>", saliencyMap, compression_params);
					imwrite("<INSERT IMAGE PATH>", saliencyBinary, compression_params);
					imwrite("<INSERT IMAGE PATH>", flowMagnitude, compression_params);
					imwrite("<INSERT IMAGE PATH>", flowProcessResult, compression_params);

					cout << "Images stored" << endl;
			}
		}
	}

	return 0;
}
